<?
/*
	MQTT Server for IP-Symcon 2018 by sokkederheld
	Version: */ define('VERSION', '0.20'); /*
		
	Please report any bugs you encounter in the IP-Symcon Community Forum thread.
	
	How to set up:
	- paste the code into an empty script in an appropriate place
	- run the script once, check the script output below.

	What's created during setup?
	- Server socket on port 1883, which is the standard for unencrypted MQTT traffic
	- A RegisterVariable which calls this script whenever data is received from a
		client (put in the same place as the script)
	- "Topics Root" category, under which the MQTT topics structure will be maintained
		(see below, put in the same place as the script)
	- "Sessions" category, under which the Sessions will be saved in the form of
		variables. (also put in the same place as the script)
		
	The aforementioned objects can be moved around if you want. They are refecenced
	by their ID in this script using an include file generated by the setup.
		
	How to use the MQTT Server once it's installed?
	- Connect over port 1883 and use as you would any other MQTT Server - subscribe
		to topics, publish messages and so on.
	- When an MQTT client publishes to a topic, the Server will create objects to
		mirror the topic strucutre inside the "Topics Root" category.
		This is usually done in the following way:
		- Any topic level with children gets created as a category of the same name.
		- Any topic level without children (the last part of the topic string) gets
			created as either a script or a variable of the same name, depending on the
			"retain" flag of the publish package.
		- Any topic to which a retained message is published gets created as a
			variable, as the behavior of a variable in IP-Symcon is the closest match
			to a retained message in MQTT. When a retained message is removed in MQTT,
			the variable will be set to hidden in IP-Symcon.
		- Any topic to which a non-retained message is published gets created as a
			script, which is then called with the message in the $_IPS['MESSAGE'] field.
		- If a script (or link to a script) already exists in the structure with the
			specified name at the specified path, it will get used and no new script
			will be created. The message gets passed in the $_IPS['MESSAGE'] field.
		- Same goes for an existing variable or a link to an existing variable -  if
			one exists with the specified name in the specified path, it gets used
			rather than a new variable being created.
		- The Server will guess the appropriate data type when it creates a variable.
			If later on a message is received that does not fit the variable's data
			type, the existing variable will be renamed and a new one will be created
			of a more versatile data type (i.e. a string may also hold a number, but
			not vice versa). To be on the safe side you may wish to create variables
			yourself rather than having them auto-created by the Server.
		- It's possible and encouraged to manually edit the categories, variables
			and scripts under the "Topics Root" category. It's also possible to set up
			links to variables and scripts that may reside outside of the whole topics
			structure. That way, for example, a readout of an existing sensor can be
			easily published over MQTT.
	- Publishing from IP-Symcon to MQTT can be achieved by either updating one of
		the variables residing (or linked to) in the topics structure (which would
		set retained messages) or by calling this script with the 'MESSAGE' and
		'TOPIC' fields populated (using IPS_RunScriptEx).
	
*/
$debug = false; // set to true to enable (noisy!) debug output
$uninstall = false; // !!! set to true to uninstall MQTT Server - CAUTION !!!

// the MQTT package types
define('PT_RESERVED_0',  '0');
define('PT_CONNECT',     '1');
define('PT_CONNACK',     '2');
define('PT_PUBLISH',     '3');
define('PT_PUBACK',      '4');
define('PT_PUBREC',      '5');
define('PT_PUBREL',      '6');
define('PT_PUBCOMP',     '7');
define('PT_SUBSCRIBE',   '8');
define('PT_SUBACK',      '9');
define('PT_UNSUBSCRIBE', '10');
define('PT_UNSUBACK',    '11');
define('PT_PINGREQ',     '12');
define('PT_PINGRESP',    '13');
define('PT_DISCONNECT',  '14');
define('PT_RESERVED_15', '15');

// PT_CONNECT >= PT_CONNACK return codes
define('RC_CONNECTION_SUCCESS',               '0');
define('RC_CONNECTION_FAILED_PROT_VERSION',   '1');
define('RC_CONNECTION_FAILED_ID_REJECTED',    '2');
define('RC_CONNECTION_FAILED_UNAVAILABLE',    '3');
define('RC_CONNECTION_FAILED_BAD_USER_PASS',  '4');
define('RC_CONNECTION_FAILED_NOT_AUTHORIZED', '5');

// flags usable in user config
define('NO_TYPE_ESCALATION',        1);
define('USE_VAR_PROFILE',           2);
define('NO_AUTO_INSTANCE_CREATION', 4);

// pretty names to display for the package types
$pt_names = array(
	"Reserved 0",
	"CONNECT",
	"CONNACK",
	"PUBLISH",
	"PUBACK",
	"PUBREC",
	"PUBREL",
	"PUBCOMP",
	"SUBSCRIBE",
	"SUBACK",
	"UNSUBSCRIBE",
	"UNSUBACK",
	"PINGREQ",
	"PINGRESP",
	"DISCONNECT",
	"Reserved 15"
);

$qos_level_profile_name = 'MQTT_QOS_Level';

$regvar_module_id = '{F3855B3C-7CD6-47CA-97AB-E66D346C037F}';
$ssck_module_id =   '{8062CF2B-600E-41D6-AD4B-1BA66C32D6ED}';
$dummy_module_id =  '{485D0419-BE97-4548-AA9C-C083EB82E61E}';
		
// these variables will hold the ids of objects required for operation
$regvar_id = false; // register variable for the server socket (the latter is accessed solely over the regvar)
$sessions_cat_id = false; // category in which session variables are kept
$topics_cat_id = false; // category holding the MQTT topics structure
$update_var_events_timer_id = false; // timer event that triggers update of variable "on update" events

// if a config file exists, include it here
$cfg_file_name = $_IPS['SELF'] . '_config.php';
if(file_exists($cfg_file_name)) include $cfg_file_name;

// if a config file exists, include it here
$user_cfg_file_name = $_IPS['SELF'] . '_user_config.php';
if(file_exists($user_cfg_file_name)) include $user_cfg_file_name;

// events cache
$event_ids = false;
$events = array();

// always output stuff when executing in console (for setup)
if($_IPS['SENDER'] == 'Execute') $debug = true;
debug_log('MQTT Server script started by ' . $_IPS['SENDER']);

// now we check how the script was called and take the appropriate actions!
if($_IPS['SENDER'] == 'RegisterVariable') { // if script called by regvar, handle new data
	// read buffer and convert it into assoc array
	$buf_str = RegVar_GetBuffer($regvar_id);
	if($buf_str == "") $buf_str = "{}";
	$buf_assoc = json_decode($buf_str, true);
	if($buf_assoc == null) $buf_assoc = array();
	
	// create an identifying string for the connection that sent the data
	$conn_id = $_IPS['CLIENTIP'] . ":" . $_IPS['CLIENTPORT'];
	
	if(array_key_exists($conn_id, $buf_assoc)) { // if there is data buffered for this connection
		$buf_str = hex2bin($buf_assoc[$conn_id]["content"]); // get buffer content
		$ts = $buf_assoc[$conn_id]["ts"]; // get timestamp
		if($ts > time() + 3) { // if timestamp older than 3 seconds, discard buffered data!
			unset($buf_assoc[$conn_id]);
			$buf_str = "";
		}
	} else { // if no data is buffered start with an empty buffer
		$buf_str = "";
	}
	$buf_str .= $_IPS['VALUE']; // append new data to buffer
	$buf_assoc[$conn_id]["content"] = bin2hex($buf_str); // save data in the global buffer array
	$buf_assoc[$conn_id]["ts"] = time(); // save timestamp also
	
	// try to handle the incoming data as mqtt package(s) and if successful, write the
	// remaining data (if any) into the buffer, or delete the buffer entry if no
	// data is left
	while(handle_incoming_data($buf_str, $conn_id)) { // if data could be handled
		if($buf_str != "") { // if there is still data in the buffer
			$buf_assoc[$conn_id]["content"] = bin2hex($buf_str); // write it into the global buffer array
		} else if(array_key_exists($conn_id, $buf_assoc)) { // otherwise, if no data left but buffer entry exists
			unset($buf_assoc[$conn_id]); // delete buffer entry for this connection
			break;
		}
	}
	
	// ensure that buffer content for dropped connections expires
	foreach($buf_assoc as $this_conn_id => $this_buf_assoc) {
		if(!array_key_exists("ts", $this_buf_assoc) || $this_buf_assoc["ts"] < time() + 10) {
			unset($buf_assoc[$this_conn_id]); // delete buffer entry for this connection
		}
	}
	
	// at the end convert buffer assoc array into json string and store it
	// in the RegVar's actual buffer
	$buf_str = json_encode($buf_assoc);
	RegVar_SetBuffer($regvar_id, $buf_str);
} else if($_IPS['SENDER'] == 'Variable') { // called by retained message variable
	handle_variable_update($_IPS['EVENT'], $_IPS['VARIABLE'], $_IPS['VALUE']);
	//IPS_LogMessage("MQTT Test", print_r($_IPS, true));
} else if($_IPS['SENDER'] == 'RunScript') { // called by another script in order to publish
	if(array_key_exists('TOPIC', $_IPS)) {
		$topic_name = $_IPS['TOPIC'];
	} else {
		throw new Exception('TOPIC field is missing');
	}
	if(array_key_exists('MESSAGE', $_IPS)) {
		$message = $_IPS['MESSAGE'];
	} else {
		throw new Exception('MESSAGE field is missing');
	}
	if(array_key_exists('RETAIN', $_IPS)) {
		$retain = $_IPS['RETAIN'];
	} else {
		$retain = false;
	}
	if(array_key_exists('CLIENT', $_IPS)) {
		$client_id = $_IPS['CLIENT'];
	} else {
		$client_id = false;
	}
	publish($topic_name, $message, $retain, $client_id);
} else if($_IPS['SENDER'] == 'TimerEvent') { // script timer - update variable events
	if($_IPS['EVENT'] == $update_var_events_timer_id) {
		setup_var_events();
	} else {
		process_package_timeouts();
	}
} else { // executed in console - setup!
	if(!$uninstall) {
		IPS_SetName($_IPS['SELF'], 'MQTT Server ' . VERSION . ' Core');
				
		debug_log("Setting up timer events...");
		IPS_SetScriptTimer($_IPS['SELF'], 15); // package timeouts and resends checked every 15 seconds
		
		// the following event is used to periodically check for added or changed
		// variables in the topics root category, and to update their on update events
		if($update_var_events_timer_id === false || !IPS_EventExists($update_var_events_timer_id)) {
			$update_var_events_timer_id = IPS_CreateEvent(1);
			IPS_SetParent($update_var_events_timer_id, $_IPS['SELF']);
			IPS_SetName($update_var_events_timer_id, 'Update variable events');
			IPS_SetEventCyclic($update_var_events_timer_id, 0, 0, 0, 0, 2, 15); // every 15 minutes
			IPS_SetEventActive($update_var_events_timer_id, true);
			debug_log("Variable event update timer created.");
		}
		
		debug_log("Setting up communication instances...");
		$parent_id = IPS_GetParent($_IPS['SELF']);
		
		// check whether a register variable exists which has this script assigned to it
		$regvar_ids = IPS_GetInstanceListByModuleID($regvar_module_id);
		$regvar_id = false;
		foreach($regvar_ids as $check_regvar_id) {
			$cfg = json_decode(IPS_GetConfiguration($check_regvar_id), true);
			if($cfg['RXObjectID'] == $_IPS['SELF']) {
				debug_log("A RegisterVariable was found that has this script assigned.");
				$regvar_id = $check_regvar_id;
				break;
			}
		}
		
		// if no register variable was found, create one
		if($regvar_id === false || !IPS_InstanceExists($regvar_id)) {
			$regvar_id = IPS_CreateInstance($regvar_module_id);
			IPS_SetName($regvar_id, 'Register Variable for MQTT Server');
			IPS_SetParent($regvar_id, $parent_id);
			IPS_SetProperty($regvar_id, 'RXObjectID', $_IPS['SELF']);
			IPS_ApplyChanges($regvar_id);
			debug_log("RegisterVariable created and this script assigned.");
		}
		
		// check whether the register variable is already connected to an io instance
		$regvar_inst = IPS_GetInstance($regvar_id);
		if($regvar_inst['ConnectionID'] == 0) { // if not connected, create server socket and connect it
			$ssck_id = IPS_CreateInstance($ssck_module_id);
			IPS_SetName($ssck_id, 'Server Socket for MQTT Server');
			if(!(
					IPS_SetProperty($ssck_id, 'Port', 1883) &&
					IPS_SetProperty($ssck_id, 'Open', 1) &&
					IPS_ApplyChanges($ssck_id)
			)) {
				debug_log("Warning: Was unable to configure port 1883 and open Server Socket - ensure the port is available!");
			}
			IPS_ConnectInstance($regvar_id, $ssck_id);
			debug_log("Server Socket created, configured and connected to RegisterVariable.");
		} else {
			$ssck_id = $regvar_inst['ConnectionID'];
			debug_log("An IO instance is already connected to the RegisterVariable.");
		}
		// check whether server socket is ready
		$ssck_inst = IPS_GetInstance($ssck_id);
		if($ssck_inst['InstanceStatus'] == 102) {
			debug_log("IO instance status is okay.");
		} else {
			debug_log("IO instance doesn't seem to be working correctly!");
		}
		debug_log("Setting up categories...");
		
		// ensure that a sessions category exists
		if($sessions_cat_id !== false && IPS_CategoryExists($sessions_cat_id)) {
			debug_log("Category for sessions already exists.");
		} else {
			if($sessions_cat_id !== false) {
				debug_log("Warning, old topics sessions has been deleted - will create new one!");
			}
			$sessions_cat_id = IPS_CreateCategory();
			IPS_SetName($sessions_cat_id, 'MQTT Sessions');
			IPS_SetParent($sessions_cat_id, $parent_id);
			debug_log("Category for sessions created.");
		}
		
		// ensure that a topics category exists
		if($topics_cat_id !== false && IPS_CategoryExists($topics_cat_id)) {
			debug_log("Category for topics already exists.");
		} else {
			if($topics_cat_id !== false) {
				debug_log("Warning, old topics category has been deleted - will create new one!");
			}
			$topics_cat_id = IPS_CreateCategory();
			IPS_SetName($topics_cat_id, 'MQTT Topic Root');
			IPS_SetParent($topics_cat_id, $parent_id);
			debug_log("Category for topics created.");
		}
		
		// save object ids to include file
		file_put_contents(
			$cfg_file_name,
			'<?' . "\r" .
			'$regvar_id = ' . $regvar_id . ";\r" .
			'$sessions_cat_id = ' . $sessions_cat_id . ";\r" .
			'$topics_cat_id = ' . $topics_cat_id . ";\r" .
			'$update_var_events_timer_id = ' . $update_var_events_timer_id . ";\r" .
			'?>'
		);
		debug_log("Setting up variable profiles...");
		if(!IPS_VariableProfileExists($qos_level_profile_name)) {
			$prof = IPS_CreateVariableProfile($qos_level_profile_name, 1);
			IPS_SetVariableProfileAssociation(
				$qos_level_profile_name, 0, 'At most once (0)', '', -1
			);
			IPS_SetVariableProfileAssociation(
				$qos_level_profile_name, 1, 'At least once (1)', '', -1
			);
			IPS_SetVariableProfileAssociation(
				$qos_level_profile_name, 2, 'Exactly once (2)', '', -1
			);
			debug_log("Created variable profile for QOS level.");
		} else {
			debug_log("QOS Level profile already exists.");
		}
		
		debug_log("Setting up variable update events...");
		setup_var_events();
		
		debug_log("Setting up user configuration script...");
		$user_cfg_file_header = '<?' . "\r" .
			"/*\r\tUser configuration script for the MQTT Server - do not put any code in here\r" .
			"\tbesides the configuration structure, or the Server will stop working properly!\r" .
			"\t\r\tThe following flags can be used (combine with | to use several at once):\r\r" .
			"\t* NO_TYPE_ESCALATION        - Disable type escalation for this topic filter,\r" .
			"\t\tmeaning that if a retained message gets published that can not be represented\r" .
			"\t\tby an existing variable's data type, the Server will not create a new variable\r" .
			"\t\twith a fitting data type. Warning: Such a message might not be received and\r" .
			"\t\tpublished properly!\r" .
			"\t* USE_VAR_PROFILE           - Publish and/or receive the formatted value\r" .
			"\t\trather than the raw value for the specified topic filter\r" .
			"\t* NO_AUTO_INSTANCE_CREATION - Do not automatically create instances for\r" .
			"\t\tthis topic filter, even when a received message looks like serialized\r" .
			"\t\tdata.\r" .
			"*/\r\r";
		$separation_line_before = "// *** EDIT ONLY BELOW THIS LINE! ***\r";
		$separation_line_after = "// *** EDIT ONLY ABOVE THIS LINE! ***\r";
		if(!file_exists($user_cfg_file_name)) {
			file_put_contents(
				$user_cfg_file_name,
				$user_cfg_file_header .
				$separation_line_before .
				"\$user_rules = array(\r" .
				"\t//\"foo/bar\" => NO_TYPE_ESCALATION\r" .
				");\r" . $separation_line_after .
				'?>'
			);
			debug_log("User configuration file created!");
		} else {
			$content_old = file_get_contents($user_cfg_file_name);
			$rules_section = substr($content_old, strpos($content_old, $separation_line_before) + strlen($separation_line_before));
			$rules_section = substr($rules_section, 0, strpos($rules_section, $separation_line_after));
			$content_new = $user_cfg_file_header . $separation_line_before . $rules_section . $separation_line_after . '?>';
			
			if($content_old == $content_new) {
				debug_log("User configuration file already exists and up to date!");
			} else {
				file_put_contents($user_cfg_file_name, $content_new);
				debug_log("User configuration file updated!");
			}
		}
		
		$user_cfg_script_id = @IPS_GetScriptIDByFile($user_cfg_file_name);
		if($user_cfg_script_id === false) {
			$user_cfg_script_id = IPS_CreateScript(0);
			$old_script_file_name = IPS_GetScriptFile($user_cfg_script_id);
			IPS_SetScriptFile($user_cfg_script_id, $user_cfg_file_name);
			unlink($old_script_file_name);
			IPS_SetParent($user_cfg_script_id, $parent_id);
			IPS_SetName($user_cfg_script_id, 'MQTT User Config');
			debug_log("Script object for easy access to user configuration file created.");
		}
		
		debug_log("Setup complete.");
		
	} else { // if $uninstall
		debug_log("Removing timer events...");
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		
		if($update_var_events_timer_id !== false && IPS_EventExists($update_var_events_timer_id)) {
			if(@IPS_DeleteEvent($update_var_events_timer_id)) {
				debug_log("Variable event update timer deleted.");
			} else {
				debug_log("Unable to delete variable event update timer.");
			}
		} else {
			debug_log("No variable event update timer found.");
		}
		
		debug_log("Removing communication instances...");
		$parent_id = IPS_GetParent($_IPS['SELF']);
		
		// check whether a register variable exists which has this script assigned to it
		$regvar_ids = IPS_GetInstanceListByModuleID($regvar_module_id);
		$regvar_id = false;
		foreach($regvar_ids as $check_regvar_id) {
			$cfg = json_decode(IPS_GetConfiguration($check_regvar_id), true);
			if($cfg['RXObjectID'] == $_IPS['SELF']) {
				debug_log("A RegisterVariable was found that has this script assigned.");
				$regvar_id = $check_regvar_id;
				break;
			}
		}
		
		// if register variable was found, remove it
		if($regvar_id !== false && IPS_InstanceExists($regvar_id)) {
			// check whether the register variable is  connected to an io instance
			$regvar_inst = IPS_GetInstance($regvar_id);
			if($regvar_inst['ConnectionID'] != 0) {
				$ssck_id = $regvar_inst['ConnectionID'];
				$ssck_inst = @IPS_GetInstance($ssck_id);
				if($ssck_inst !== false && $ssck_inst['ModuleInfo']['ModuleID'] == $ssck_module_id) {
					if((
							@IPS_SetProperty($ssck_id, 'Open', 0) &&
							@IPS_ApplyChanges($ssck_id)
					)) {
						debug_log("Server Socket closed.");
						IPS_Sleep(1000);
					} else {
						debug_log("Warning: Was unable to close Server Socket.");
					}
					if(@IPS_DeleteInstance($ssck_id)) {
						debug_log("Server Socket deleted.");
					} else {
						debug_log("Unable to delete Server Socket.");
					}
				} else {
					debug_log("IO Instance is not a Server Socket! Will not delete it.");
				}
			} else {
				debug_log("No IO Instance is connected to the Register Variable.");
			}
			
			if(@IPS_DeleteInstance($regvar_id)) {
				debug_log("Register Variable deleted.");
			} else {
				debug_log("Unable to delete Register Variable.");
			}
		} else {
			debug_log("No RegisterVariable was found that has this script assigned.");
		}
			
		debug_log("Removing categories...");
		
		// ensure that a sessions category exists
		if($sessions_cat_id !== false && IPS_CategoryExists($sessions_cat_id)) {
			$session_var_ids = IPS_GetChildrenIDs($sessions_cat_id);
			foreach($session_var_ids as $session_var_id) {
				$session_name = IPS_GetName($session_var_id);
				if(delete_session_variable($session_var_id)) {
					debug_log("Session variable \"" . $session_name . "\" deleted.");
				}
			}
			if(@IPS_DeleteCategory($sessions_cat_id)) {
				debug_log("Category for sessions deleted.");
			} else {
				debug_log("Unable to delete category for sessions.");
			}
		} else {
			debug_log("Category for sessions was not found.");
		}
		
		// ensure that a topics category exists
		if($topics_cat_id !== false && IPS_CategoryExists($topics_cat_id)) {
			if(@IPS_DeleteCategory($topics_cat_id)) {
				debug_log("Category for topics deleted.");
			} else {
				debug_log("Category for topics could not be deleted. If it's not empty, please delete it manually!");
			}
		} else {
			debug_log("Category for topics was not found.");
		}
		
		if(file_exists($cfg_file_name)) {
			if(unlink($cfg_file_name)) {
				debug_log("Configuration file deleted.");
			} else {
				debug_log("Unable to delete configuration file \"" . $cfg_file_name . "\".");
			}
		} else {
			debug_log("No configuration file was found.");
		}
		
		debug_log("Removing variable profiles...");
		if(IPS_VariableProfileExists($qos_level_profile_name)) {
			if(IPS_DeleteVariableProfile($qos_level_profile_name)) {
				debug_log("Variable profile for QOS level deleted.");
			} else {
				debug_log("Variable profile for QOS level could not be deleted.");
			}
		} else {
			debug_log("QOS Level profile was not found.");
		}

		debug_log("Removing events...");
		$children_ids = IPS_GetChildrenIDs($_IPS['SELF']);
		foreach($children_ids as $child_id) {
			$name = IPS_GetName($child_id);
			if(@IPS_DeleteEvent($child_id)) {
				debug_log("Event \"" . $name . "\" deleted.");
			} else {
				debug_log("Unable to delete \"" . $name . "\".");
			}
		}
		
		debug_log("Uninstall complete.");
	}
}

// Handle incoming data as mqtt package if possible.
// Returns true if parsing completed and data should be removed from the buffer,
// false if data was incomplete and more data is expected. A result of true only
// means that the parsing didn't lack any data, not that the package led to any
// meaningful action or that the request was successful in any way!
// $data is the buffer content and gets modified by this function. If true is
// returned, the new value of $data should be stored in the RegVar's buffer
// structure for the current $conn_id.
function handle_incoming_data(&$data, $conn_id) {
	global $debug, $pt_names;
	// first, try parsing the MQTT package header
	try {
		$header = extract_header($data);
		if($header === false) return false; // incomplete data
	} catch(Exception $e) { // data complete, but something's wrong with the header
		debug_log("Error parsing header: " . $e->getMessage());
		debug_log("Data: " . chunk_split(bin2hex($data), 2, ' '));
		$data = '';
		return true;
	}
	
	// check whether to expect a payload
	$expect_payload = false;
	switch($header['type']) {
		case PT_PUBLISH:
			if($header['remaining_len'] == 0) break; // for this package type, payload is optional
		case PT_CONNECT:
		case PT_SUBSCRIBE:
		case PT_UNSUBSCRIBE:
			$expect_payload = true;
	}
	
	// try parsing the payload if one is expected
	if($expect_payload) {
		try {
			$payload = extract_payload($header, $data);
			if($payload === false) return false; // incomplete data
		} catch(Exception $e) { // data complete, but something's wrong with the payload
			debug_log("Error parsing payload: " . $e->getMessage());
			$data = '';
			return true;
		}
	}
	
	// at this point, remaining_len must be zero. If it's not, there must be some
	// error on the sending or the receiving end.
	if($header['remaining_len'] != 0) {
		debug_log('Warning: Package indicates ' . $header['remaining_len'] . ' additional byte(s) which were not expected by the parser.');
	}
	
	if($debug) {
		debug_log("Complete package of type " . $pt_names[$header['type']] . " parsed.");
		debug_log("Header: " . print_r($header, true));
		if($expect_payload) debug_log("Payload: " . print_r($payload, true));
	}
	
	// flags that control whether the session must be updated (new timestamp and/or
	// session data written) or if it ends (disconnect). If the session is a per-
	// manent one, it doesn't really end, but actually the conn_id is just cleared
	// so that another connection can resume the session!
	$update_session = false;
	$end_session = false;
	
	// if all data is parsed without an exception occured, we can now process it
	if($header['type'] == PT_CONNECT) {
		//IPS_LogMessage('MQTT', 'Connect: ' . $payload['client_id']);
		$return_code = RC_CONNECTION_SUCCESS;
				
		// level 3 (MQTT 3.1) and 4 (MQTT 3.1.1) are supported. Check for valid
		// protocol level / protocol name combination
		if($header['protocol_level'] == 4) {
			if(strtoupper($header['protocol_name']) != 'MQTT') {
				$return_code = RC_CONNECTION_FAILED_PROT_VERSION;
			} else {
				// if no client id is supplied, create one
				if(strlen($payload['client_id']) == 0) {
					if(!$header['connect_flags']['clean_session']) { // works only for "clean" session
						$return_code = RC_CONNECTION_FAILED_ID_REJECTED;
					}
					// create a temporary client id and paste it into the payload assoc array
					$payload['client_id'] = conn_id_to_ident($conn_id) . 't' . date("Ymdhis");
				}
			}
		} else if($header['protocol_level'] == 3) {
			if(strtoupper($header['protocol_name']) != 'MQISDP') {
				$return_code = RC_CONNECTION_FAILED_PROT_VERSION;
			} else {
				if(strlen($payload['client_id']) == 0 || strlen($payload['client_id']) > 23) {
					$return_code = RC_CONNECTION_FAILED_ID_REJECTED;
				}
			}
		} else {
			$return_code = RC_CONNECTION_FAILED_PROT_VERSION;
		}
		
		// retrieve the session's assoc array
		$session_assoc = get_session(
			$conn_id,
			$payload['client_id'],
			true, // create session if none exists
			$header['connect_flags']['clean_session']
		);
		
		// store keep-alive timeout in the assoc array
		$session_assoc['kal'] = $header['kal'];
		$session_assoc['protocol_level'] = $header['protocol_level'];
		
		// if a will is supplied, store it in the session assoc array
		if($header['connect_flags']['will_flag']) {
			$session_assoc['will_flag'] = true;
			$session_assoc['will_retain'] = $header['connect_flags']['will_retain'];
			$session_assoc['will_qos'] = $header['connect_flags']['will_qos'];
			$session_assoc['will_topic'] = $payload['will_topic'];
			$session_assoc['will_message'] = $payload['will_message'];
		} else { // if no will supplied, remove all will data from assoc array
			$session_assoc['will_flag'] = false;
			unset($session_assoc['will_retain']);
			unset($session_assoc['will_qos']);
			unset($session_assoc['will_topic']);
			unset($session_assoc['will_message']);
		}
		
		// if the connection attempt succeeded, set the appropriate flags, otherwise
		// end the session (only MQTT 3.1.1 supports the "session exists" flag!)
		if($return_code == RC_CONNECTION_SUCCESS) {
			$ca_flags = $header['protocol_level'] == 3 || $session_assoc['new'] ?
				0b00000000 : 0b00000001;
			$update_session = true;
		} else {
			$ca_flags = 0b00000000;
			$end_session = true;
		}
		
		// try to send the PT_CONNACK package. If it fails, end the session.
		if(!send_package($conn_id, PT_CONNACK, 0b0000, array($ca_flags, $return_code))) {
			$update_session = false;
			$end_session = true;
		}
	} else { // all other packages than PT_CONNECT
		$session_assoc = get_session($conn_id);
		if($session_assoc !== false) { // if there is a session for this connection id
			// by default, update the session (resetting the keep alive timeout timer)
			$update_session = true;
			$end_session = false;
			
			// handle package depending on the type
			switch($header['type']) {
				case PT_PINGREQ: // ping request - just send a response
					send_package($conn_id, PT_PINGRESP, 0b0000);
					break;
					
				case PT_DISCONNECT: // disconnect - discard last will and end session gracefully
					$session_assoc['will_flag'] = false;
					unset($session_assoc['will_retain']);
					unset($session_assoc['will_qos']);
					unset($session_assoc['will_topic']);
					unset($session_assoc['will_message']);
					$end_session = true;
					$update_session = false;
					break;
					
				case PT_PUBLISH: // publish - publish message to all subscribers, store
					// message in a variable if it's a retained message
					
					$publish_now = true;
					if($header['flags']['qos'] == 1) { // at least once
						send_package($conn_id, PT_PUBACK, 0b0000, array(
							msb($header['package_id']), lsb($header['package_id'])
						));
					} else if($header['flags']['qos'] == 2) { // exactly once
						if(array_key_exists($header['package_id'], // if package was received before, do not republish to clients
							$session_assoc['package_ids']['incoming_received_qos_2'])) {
							$publish_now = false;
						}
						$session_assoc['package_ids']['incoming_received_qos_2'][$header['package_id']] = time();
						send_package($conn_id, PT_PUBREC, 0b0000, array(
							msb($header['package_id']), lsb($header['package_id'])
						));
						// awaiting PT_PUBREL now
					}
					
					// publish now (unless we already did so before)
					if($publish_now){ // && $expect_payload) {
						try {
							publish(
								$header['topic_name'],
								$expect_payload ? preg_replace('/[[:^print:]]/', "", $payload) : "",
								$header['flags']['retain'],
								$session_assoc['client_id']
							);
						} catch(Exception $e) { // if publish failed (forbidden to write variable) end session
							debug_log($e->getMessage());
							$end_session = true; // not very elegant, but there is no "negative ack" for this in MQTT
						}
					}
					break;
					
				case PT_PUBREL: // the client's response to PUBREC, completes incoming qos 2
					if(array_key_exists($header['package_id'],
						$session_assoc['package_ids']['incoming_received_qos_2'])) {
						send_package($conn_id, PT_PUBCOMP, 0b0000, array(
							msb($header['package_id']), lsb($header['package_id'])
						));
						unset($session_assoc['package_ids']['incoming_received_qos_2'][$header['package_id']]);
					}
					break;
				
				case PT_PUBACK: // the client's response to PUBLISH with qos 1
					if(array_key_exists($header['package_id'],
						$session_assoc['package_ids']['outgoing_sent_qos_1'])) {
						unset($session_assoc['package_ids']['outgoing_sent_qos_1'][$header['package_id']]);
					}
					break;
					
				case PT_PUBREC: // the client's response to PUBLISH with qos 2
					send_package($conn_id, PT_PUBREL, 0b0000, array(
						msb($header['package_id']), lsb($header['package_id'])
					));
					if(array_key_exists($header['package_id'],
						$session_assoc['package_ids']['outgoing_sent_qos_2'])) {
						// save package in "outgoing received qos 2" list and remove from "sent qos 2" list
						$session_assoc['package_ids']['outgoing_received_qos_2'][$header['package_id']] =
							$session_assoc['package_ids']['outgoing_sent_qos_2'][$header['package_id']];
						unset($session_assoc['package_ids']['outgoing_sent_qos_2'][$header['package_id']]);
						// awaiting PUBCOMP now
					}
					break;
					
				case PT_PUBCOMP: // the client's response to PUBREL
					if(array_key_exists($header['package_id'],
						$session_assoc['package_ids']['outgoing_received_qos_2'])) {
						unset($session_assoc['package_ids']['outgoing_received_qos_2'][$header['package_id']]);
					}
					break;
					
				case PT_SUBSCRIBE: // subscribe - subscribe to one or more topics/topic filters
					if(count($payload) == 0) { // at least one topic filter must be specified
						$end_session = true;
						$update_session = false;
						break;
					}
					$bytes = array( // all subscribe packages have a package id, so send it back too
						msb($header['package_id']), lsb($header['package_id'])
					);
					// for each topic filter, perform the subscription and push the return code
					// to the return bytes array (code == qos level of the corresponding topic
					// filter, each code in the order of the supplied filter strings!)
					for($i = 0; $i < count($payload); $i++) {
						array_push($bytes,
							subscribe($session_assoc['client_id'], $payload[$i]['topic_filter'], $payload[$i]['qos'])
						);
					}
					send_package($conn_id, PT_SUBACK, 0b0000, $bytes);
					break;
				
				case PT_UNSUBSCRIBE: // unsubscribe from topics / filters
					if(count($payload) == 0) { // at least one topic filter must be specified
						$end_session = true;
						$update_session = false;
						break;
					}
					$bytes = array( // all unsubscribe packages have a package id, so send it back too
						msb($header['package_id']), lsb($header['package_id'])
					);
					for($i = 0; $i < count($payload); $i++) {
						unsubscribe($session_assoc['client_id'], $payload[$i]);
					}
					send_package($conn_id, PT_UNSUBACK, 0b0000, $bytes);
					break;
			}
		} else {
			debug_log('No session!');
		}
	}
	
	if($end_session) {
		end_session($session_assoc);
	} else if($update_session) {
		update_session($session_assoc);
	}
	return true;
} // handle_incoming_data

// process package timeouts and do pending resends for all active client sessions
function process_package_timeouts() {
	$sessions = get_active_client_sessions();
	foreach($sessions as $session_assoc) {
		process_package_timeouts_for_session($session_assoc);
	}
} // process_package_timeouts

// process the package timeouts and -resends for the specified session
function process_package_timeouts_for_session($session_assoc) {
	$timeout = 30;
	
	// PUBREL not received? resend PUBREC
	foreach($session_assoc['package_ids']['incoming_received_qos_2'] as $package_id => $sent_time) {
		if($sent_time + $timeout > time()) {
			if(!send_package($session_assoc['connection_id'], PT_PUBREC, 0b0000, array(
				msb($package_id), lsb($package_id)
			))) {
				end_session($session_assoc);
				return;
			}
		}
	}
	
	// PUBACK not received? resend PUBLISH
	foreach($session_assoc['package_ids']['outgoing_sent_qos_1'] as $package_id => $package_info) {
		if($package_info['sent'] + $timeout > time()) {
			if(!send_publish_package(
				$package_info['topic'], $package_info['message'],
				$session_assoc['connection_id'],
				1, // qos
				$session_assoc,
				false, // retain
				$package_id
			)) {
				end_session($session_assoc);
				return;
			}
		}
	}
	
	// PUBREC not received, resend PUBLISH
	foreach($session_assoc['package_ids']['outgoing_sent_qos_2'] as $package_id => $package_info) {
		if($package_info['sent'] + $timeout > time()) {
			if(!send_publish_package(
				$package_info['topic'], $package_info['message'],
				$session_assoc['connection_id'],
				2, // qos
				$session_assoc,
				false, // retain
				$package_id
			)) {
				end_session($session_assoc);
				return;
			}
		}
	}
	
	// PUBCOMP not received, resend PUBREL
	foreach($session_assoc['package_ids']['outgoing_received_qos_2'] as $package_id => $package_info) {
		if($package_info['sent'] + $timeout > time()) {
			if(!send_package($session_assoc['connection_id'], PT_PUBREL, 0b0000, array(
				msb($package_id), lsb($package_id)
			))) {
				end_session($session_assoc);
				return;
			}
		}
	}
	
} // process_package_timeouts_for_session

// return an array of all active client sessions as assoc arrays
function get_active_client_sessions() {
	global $sessions_cat_id;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	$sessions = array();
	$session_var_ids = IPS_GetChildrenIDs($sessions_cat_id);
	if($session_var_ids === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category was deleted - run setup!');
	}
	foreach($session_var_ids as $var_id) {
		$conn_id = ident_to_conn_id(IPS_GetObject($var_id)['ObjectIdent']);
		if($conn_id != '') {
			$client_id = IPS_GetName($var_id);
			$session_assoc = get_session($conn_id, $client_id);
			if($session_assoc !== false) {
				$kal = $session_assoc['kal'];
				$last_update = $session_assoc['last_update'];
				if($kal == 0 || time() <= $last_update + $kal * 1.5) {
					$sessions[] = $session_assoc;
				} else { // don't return timed out session
					perform_last_will($session_assoc);
					end_session($session_assoc);
				}
			}
		}
	}
	
	return $sessions;
} // get_active_client_sessions

function perform_last_will($session_assoc) {
	if($session_assoc['will_flag']) {
		publish(
			$session_assoc['will_topic'],
			$session_assoc['will_message'],
			$session_assoc['will_retain'],
			$session_assoc['client_id'],
			$session_assoc['will_qos'] // this will not be honored for retained will messages
		);
		$session_assoc['will_flag'] = false;
		unset($session_assoc['will_retain']);
		unset($session_assoc['will_qos']);
		unset($session_assoc['will_topic']);
		unset($session_assoc['will_message']);
		update_session($session_assoc);
	}
} // perform_last_will

// Sets up "on update" events for all variables within the "topics" structure.
// This must be called after any changes are made to the topics structure in order
// for all the "retained message" variables to get published correctly to any
// subscribers!
function setup_var_events($parent_id = false) {
	global $topics_cat_id, $event_ids, $events;
	if($topics_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Topics Root category not configured - run setup!');
	}
	
	$clear_cache = false;
	if($parent_id === false) {
		$parent_id = $topics_cat_id;
		$clear_cache = true; // clear the cache in the end if this is not a recursive call
	}
	
	// get a list of sub topics under the current category id
	$sub_topics = IPS_GetChildrenIDs($parent_id);
	foreach($sub_topics as $sub_topic_id_original) {
		$obj = IPS_GetObject($sub_topic_id_original);
		$sub_topic_id = $sub_topic_id_original;
		$linked = false;
		
		if($obj['ObjectType'] == 6) { // if topic object is link, follow it
			$sub_topic_id = IPS_GetLink($sub_topic_id)['TargetID'];
			$obj = IPS_GetObject($sub_topic_id);
			$linked = true;
		}
		
		if($obj['ObjectType'] == 2) { // if topic object is variable
			if(!is_shadow_var($sub_topic_id) && !is_vestigial_var($sub_topic_id)) {
				$event_id = ensure_var_event($sub_topic_id); // create variable event if not existant
				IPS_SetName($event_id, get_topic_from_id($sub_topic_id_original));
			}
		} else if($obj['ObjectType'] == 0 && !$linked) { // if topic object is category
			setup_var_events($sub_topic_id); // recursively call for sub topic
		} else if($obj['ObjectType'] == 3) { // script
		} else if($obj['ObjectType'] == 1 && !$linked) { // instance
		} else { // unsupported object type or unsupported linking
			debug_log("Warning! Unsupported object in topics structure: ID " . $sub_topic_id_original);
		}
	}
	
	// get a list of all events calling this script, but only if the list doesn't
	// already exist in the cache
	if($event_ids === false) $event_ids = IPS_GetChildrenIDs($_IPS['SELF']);
	foreach($event_ids as $event_id) {
		if(array_key_exists($event_id, $events)) {
			$event = $events[$event_id];
		} else {
			$event = IPS_GetEvent($event_id);
			$events[$event_id] = $event;
		}
		// if the triggering variable of an event has been deleted, delete the event
		if(!IPS_ObjectExists($event['TriggerVariableID'])) {
			@IPS_DeleteEvent($event_id);
		} else if(
			is_shadow_var($event['TriggerVariableID']) ||
			is_vestigial_var($event['TriggerVariableID'])
		) {
			@IPS_DeleteEvent($event_id);
		}
		
		// in case the target of a link changes, the event will be
		// deleted in handle_variable_update() when the event is raised!
	}
	
	if($clear_cache) {
		// reset the cache
		$event_ids = false;
		$events = array();
	}
} // setup_var_events

// handler to be called when an "on update" event occurs for a variable
function handle_variable_update($event_id, $var_id, $message) {
	// do not handle variable events for shadow variables
	if(is_shadow_var($var_id) || is_vestigial_var($var_id)) return;
	
	// if the variable resides within the topics structure, we can get the topic
	// name from this function
	$topic_name = get_topic_from_id($var_id);
		
	// if the topic name has not been found, we try to derive the topic name
	// from the event name. This may be the case for linked variables, which do
	// not reside within the topics structure directly (only a link to them)
	if($topic_name === false) {
		$event_name = @IPS_GetName($event_id);
		if($event_name !== false && strpos($event_name, '\\') === false) {
			$topic_name = $event_name;
			$link_id = get_topic_link_id($topic_name);
			if($link_id !== false) {
				$link = IPS_GetLink($link_id);
				$target_id = $link['TargetID'];
				if($target_id != $var_id) { // link target changed? delete this old event!
					@IPS_DeleteEvent($event_id);
					return;
				}
			}
		}
	}
	
	// if the topic name could be determined
	if($topic_name !== false) {
		$flags = get_user_flags($topic_name);
		
		if($flags & USE_VAR_PROFILE) {
			$message = GetValueFormatted($var_id);
		} else if(is_bool($message)) {
			$message = $message ? "true" : "false";
		}
		
		if($message !== '') {
			IPS_SetHidden($var_id, false);
			send_publish_packages($topic_name, $message);
		} else {
			IPS_SetHidden($var_id, true);
		}
	}
} // handle_variable_update

// returns true if the variable with the given id is a shadow variable.
// Shadow variables can be created for topics that are expected to reveive
// non-retained messages, in order to hold the topic's last message permanently
// for use within IPS (only). A shadow variable is marked by its "*" prefix.
function is_shadow_var($var_id) {
	$name = @IPS_GetName($var_id);
	if($name === false) return false;
	return substr($name, 0, 1) == '*';
} // is_shadow_var

function is_vestigial_var($var_id) {
	$name = @IPS_GetName($var_id);
	if($name === false) return false;
	return strpos($name, '@') !== false;
} // is_vestigial_var

// return any flags set in the user configuration file for the specified topic
function get_user_flags($topic_name) {
	global $user_rules;
	if(!isset($user_rules)) return 0;
	
	try {
		$result = 0;
		foreach($user_rules as $filter => $flags) {
			if(topic_matches_filter($topic_name, $filter)) {
				$result |= $flags;
			}
		}
	} catch(Exception $e) {
		debug_log("Error applying user defined rules: " . $e->getMessage());
	}
	return $result;
} // get_user_flags

// send around publish packages to all subscribers of the specified topic name
function send_publish_packages($topic_name, $message, $qos_override = false) {
	$subscribers = get_subscribed_clients($topic_name);
	foreach($subscribers as $subscriber) {
		if($qos_override !== false) {
			$qos = $qos_override;
		} else {
			$qos = $subscriber['qos'];
		}
		send_publish_package(
			$topic_name, $message,
			$subscriber['conn_id'],
			$qos,
			$qos > 0 ? // only if qos > 0 get and provide session
				get_session($subscriber['conn_id'], $subscriber['client_id']) :
				false // session is not needed for qos 0 (fire and forget)
		);
	}
} // send_publish_packages

// send publish package for specified topic name / message over the specified
// $conn_id with the specified $qos level
function send_publish_package(
	$topic_name,
	$message,
	$conn_id,
	$qos,
	$session_assoc = false,
	$retain = false,
	$package_id = false
) {
	$flags = 0b0000;
	if($qos == 1) { 
		$flags = $flags | 0b0010;
	} else if($qos == 2) {
		$flags = $flags | 0b0100;
	}
	if($retain) { // set retain flag if package sent due to new subscription!
		$flags = $flags | 0b0001;
	}
	$bytes = array();
	
	$bytes[] = msb(strlen($topic_name));
	$bytes[] = lsb(strlen($topic_name));
	for($i = 0; $i < strlen($topic_name); $i++) {
		$bytes[] = ord($topic_name[$i]);
	}
	
	// if qos > 0 is used, provide package id and store it in session
	if($qos > 0) {
		if($package_id === false) {
			$package_id = $session_assoc['package_ids']['outgoing_next'];
			$next_id = $package_id + 1;
			if($next_id > 65535) $next_id = 1;
			$session_assoc['package_ids']['outgoing_next'] = $next_id;
		} else {
			$flags = $flags | 0b1000; // the ever useless DUP flag gets set
		}
		// store topic as well as message for the package id, so the package can be
		// resent if it doesn't get confirmed by the client
		$session_assoc['package_ids']['outgoing_sent_qos_' . $qos][$package_id] = array(
			'topic' => $topic_name,
			'message' => $message,
			'sent' => time()
		);
		$bytes[] = msb($package_id);
		$bytes[] = lsb($package_id);
		update_session($session_assoc);
	}
	
	$message = sane_string_cast($message); // convert message to string so the byte encoding will work properly
	for($i = 0; $i < strlen($message); $i++) {
		$bytes[] = ord($message[$i]);
	}
	
	if($conn_id != '') {
		return send_package($conn_id, PT_PUBLISH, $flags, $bytes);
	}
	return false;
} // send_publish_package

// subscribe the specified client session to the specified topic filter using the
// specified qos level. If there is a retained message matching the filter, it is
// published to the subscriber immediately.
function subscribe($client_id, $topic_filter, $qos) {
	global $sessions_cat_id, $qos_level_profile_name;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	$session_var_id = @IPS_GetObjectIDByName($client_id, $sessions_cat_id);
	if($session_var_id === false) throw new Exception("Client ID has no session: \"" . $client_id . "\"");
	
	$retain = false;
	
	// if a subscription variable already exists for the specified filter, reuse it
	$subscription_var_id = @IPS_GetVariableIDByName($topic_filter, $session_var_id);
	if($subscription_var_id === false) { // if necessary, create subscription var
		$subscription_var_id = IPS_CreateVariable(1);
		IPS_SetParent($subscription_var_id, $session_var_id);
		IPS_SetName($subscription_var_id, $topic_filter);
		IPS_SetVariableCustomProfile($subscription_var_id, $qos_level_profile_name);
		$retain = true;
	}
	SetValue($subscription_var_id, $qos); // value of subscription var is the qos level
	
	// get array of all ids of variables which contain retained messages for the
	// specified topic filter
	$retained_var_list = get_retained_var_list($topic_filter);
	if(count($retained_var_list) > 0) $conn_id = ident_to_conn_id(
		IPS_GetObject($session_var_id)['ObjectIdent']
	);
	foreach($retained_var_list as $retained_var_list_entry) {
		send_publish_package(
			$retained_var_list_entry['topic_name'],
			GetValue($retained_var_list_entry['variable_id']), // message
			$conn_id,
			$qos,
			$qos > 0 ? // session is needed only for qos > 0
				get_session($conn_id, $client_id) :
				false,
			$retain // retain flag, means that this package arrives due to a new subscription
		);
	}
	
	return $qos; // might also return 128 for failed subscriptions in later update!
} // subscribe

function unsubscribe($client_id, $topic_filter) {
	global $sessions_cat_id;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	$session_var_id = @IPS_GetObjectIDByName($client_id, $sessions_cat_id);
	if($session_var_id === false) throw new Exception("Client ID has no session: \"" . $client_id . "\"");
	
	$subscription_var_id = @IPS_GetVariableIDByName($topic_filter, $session_var_id);
	if($subscription_var_id !== false) {
		IPS_DeleteVariable($subscription_var_id);
	}
} // unsubscribe

// Returns the ids of all variables which currently hold "retained" messages and which
// match the specified topic filter. This function gets used when a client just
// subscribed, in order to deliver all retained messages due to the subscription.
function get_retained_var_list($topic_filter, $this_cat_id = false, $topic = '') {
	global $topics_cat_id;
	if($topics_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Topics Root category not configured - run setup!');
	}
	
	if($this_cat_id === false) $this_cat_id = $topics_cat_id;
	$retained_var_list = array();
	
	// unless we're at the # wildcard, go one level deeper...
	if($topic_filter != '#') {
		$slash_pos = strpos($topic_filter, '/');
		if($slash_pos !== false) {
			$this_level_filter = substr($topic_filter, 0, $slash_pos);
			$topic_filter = substr($topic_filter, $slash_pos + 1);
		} else {
			$this_level_filter = $topic_filter;
			$topic_filter = '';
		}
	} else {
		$this_level_filter = '#';
	}
	
	// recursively walk over the topics and if the current level matches, add its
	// variable ids to the result array
	$children_ids = IPS_GetChildrenIDs($this_cat_id);
	foreach($children_ids as $child_id) {
		$obj = IPS_GetObject($child_id);
		$name = IPS_GetName($child_id);
		if($name == $this_level_filter || $this_level_filter == '+' || $this_level_filter == '#') {
			if($obj['ObjectType'] == 0) { // category
				$retained_var_list = array_merge($retained_var_list, get_retained_var_list(
					$topic_filter, $child_id, $topic . $name . '/'
				));
			} else if($obj['ObjectType'] == 2) { // variable
				// if hidden, a variable exists but no retained message is stored
				if(!$obj['ObjectIsHidden'] && !is_shadow_var($child_id) && !is_vestigial_var($child_id)) {
					$retained_var_list[] = array(
						'variable_id' => $child_id,
						'topic_name' => $topic . $name
					);
				}
			} else if($obj['ObjectType'] == 6) { // link
				$target_id = IPS_GetLink($child_id)['TargetID'];
				$obj = IPS_GetObject($target_id);
				if($obj['ObjectType'] == 2) { // it target is a variable
					$retained_var_list[] = array(
						'variable_id' => $target_id,
						'topic_name' => $topic . $name
					);
				}
			}
		}
	}
	
	return $retained_var_list;
} // get_retained_var_list

// Returns an array of assoc arrays which hold informations on all clients which
// are subscribed to the specified topic name.
// Only a single topic name can be specified here, not a topic filter!
function get_subscribed_clients($topic_name) {
	global $sessions_cat_id;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	$clients = array();
	
	$session_var_ids = @IPS_GetChildrenIDs($sessions_cat_id);
	if($session_var_ids === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category was deleted - run setup!');
	}
	foreach($session_var_ids as $session_var_id) {
		$subscription_var_ids = @IPS_GetChildrenIDs($session_var_id);
		if($subscription_var_ids !== false) {
			foreach($subscription_var_ids as $subscription_var_id) {
				$topic_filter = IPS_GetName($subscription_var_id);
				if(topic_matches_filter($topic_name, $topic_filter)) {
					$qos = GetValue($subscription_var_id);
					$clients[] = array(
						'client_id' => IPS_GetName($session_var_id), // never used?
						'conn_id' => ident_to_conn_id(IPS_GetObject($session_var_id)['ObjectIdent']),
						'variable_id' => $session_var_id,
						'qos' => $qos
					);
					//break;
				}
			}
		}
	}
	
	return $clients;
} // get_subscribed_clients

// return the topic name for any element that is part of the topics structure.
// if the element is not part of the topics structure, false is returned.
function get_topic_from_id($id, $topic_name = '') {
	global $topics_cat_id;
	if($topics_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Topics Root category not configured - run setup!');
	}
	
	$topic_name = IPS_GetName($id) . $topic_name;
	if($topic_name == '/') $topic_name = '';
	$parent_id = @IPS_GetParent($id);
	if($parent_id == $topics_cat_id) {
		return $topic_name;
	} else if($parent_id != 0) {
		return get_topic_from_id($parent_id, '/' . $topic_name);
	} else {
		return false;
	}
} // get_topic_from_id

// returns true if the specified topic name matches the specified filter
function topic_matches_filter($topic_name, $topic_filter) {
	while(strlen($topic_name) > 0) {
		$slashpos = strpos($topic_name, '/');
		if($slashpos !== false) {
			$part_topic = substr($topic_name, 0, $slashpos);
			$topic_name = substr($topic_name, $slashpos + 1);
		} else {
			$part_topic = $topic_name;
			$topic_name = '';
		}
		$slashpos = strpos($topic_filter, '/');
		if($slashpos !== false) {
			$part_filter = substr($topic_filter, 0, $slashpos);
			$topic_filter = substr($topic_filter, $slashpos + 1);
		} else {
			$part_filter = $topic_filter;
			$topic_filter = '';
		}
		
		// what we actually look for is MISmatches
		if($part_topic != $part_filter) {
			// but if the strings aren't the same, the filter might still contain a
			// wildcard character at this level, so we check for that now
			if($part_filter == '#') { // '#' means "match all below this level, so no further checks
				break;
			} else if($part_filter != '+') { // not matching and not '+' wildcard - no match
				return false;
			}
		}
	}
	// if no mismatch has been found at this point, it must be a match
	return true;
} // topic_matches_filter

// publishes the supplied $message under the supplied $topic_name
function publish($topic_name, $message, $retain, $client_id = false, $qos_override = false) {
	$flags = get_user_flags($topic_name);
	$topic_var_info = get_topic_var_info($topic_name, $message, $retain);
	
	if($flags & USE_VAR_PROFILE) {
		$profile_name = get_var_profile($message, $topic_var_info['default_var_id'], $profile_name_guessed);
	}
	
	// set variable, but only for retained messages
	if($topic_var_info['default_var_id'] !== false) {
		// if there is a read only variable for this topic, it can not be published
		// to, even if the message is not set to be retained!
		$obj = IPS_GetObject($topic_var_info['default_var_id']);
		if($obj['ObjectIsReadOnly']) {
			throw new Exception('Topic ' . $topic_name . ' can not be published to: Read-only variable!');
		}
		
		if($retain) {
			IPS_SetHidden($topic_var_info['default_var_id'], $message === '');
			if($topic_var_info['vestigial_var_ids']['bool'] !== false) {
				IPS_SetHidden($topic_var_info['vestigial_var_ids']['bool'], $message === '');
			}
			if($topic_var_info['vestigial_var_ids']['int'] !== false) {
				IPS_SetHidden($topic_var_info['vestigial_var_ids']['int'], $message === '');
			}
			if($topic_var_info['vestigial_var_ids']['float'] !== false) {
				IPS_SetHidden($topic_var_info['vestigial_var_ids']['float'], $message === '');
			}
			
			$var = @IPS_GetVariable($topic_var_info['default_var_id']);
			if($var !== false) {
				if($message !== '') {// || $var['VariableType'] == 3) {
					if($flags & USE_VAR_PROFILE) {
						if($profile_name != '') {
							if(
								($var['VariableProfile'] == '' && $var['VariableCustomProfile'] == '') ||
								!$topic_var_info['default_var_is_linked']
							) {
								IPS_SetVariableCustomProfile($topic_var_info['default_var_id'], $profile_name);
							}
						}
					}
					
					$message_bool = ($message == 'false') ? false : boolval($message);
					$message_int = intval($message);
					$message_float = floatval($message);
					
					if($var['VariableType'] == 0) $message = $message_bool;
					if($var['VariableType'] == 1) $message = $message_int;
					if($var['VariableType'] == 2) $message = $message_float;
					@SetValue($topic_var_info['default_var_id'], $message);
					
					if($topic_var_info['vestigial_var_ids']['bool'] !== false) {
						@SetValue($topic_var_info['vestigial_var_ids']['bool'], $message_bool);
					}
					if($topic_var_info['vestigial_var_ids']['int'] !== false) {
						@SetValue($topic_var_info['vestigial_var_ids']['int'], $message_int);
					}
					if($topic_var_info['vestigial_var_ids']['float'] !== false) {
						@SetValue($topic_var_info['vestigial_var_ids']['float'], $message_float);
					}
				}
			}
		}
	}
	
	// if a shadow variable exists, always set it!
	if($topic_var_info['shadow_var_id'] !== false) {
		$var = @IPS_GetVariable($topic_var_info['shadow_var_id']);
		if($var !== false) {
			if(($message == 'true' || $message == 'false') && $var['VariableType'] < 3) {
				$message = ($message == 'true');
			}
			@SetValue($topic_var_info['shadow_var_id'], $message);
		}
	}
	
	// if message looks like a json object, create topic instance to hold the
	// unserialized data
	$allow_create = false;
	if(!($flags & NO_AUTO_INSTANCE_CREATION)) {
		if(substr($message, 0, 1) == '{' && substr($message, -1) == '}') {
			$allow_create = true;
		}
	}
	// if a topic instance exists, try to populate it with the data from the
	// received message (json)
	$topic_instance_id = get_topic_instance_id($topic_name, $allow_create);
	if($topic_instance_id !== false) {
		if(!populate_topic_instance($topic_instance_id, $message)) {
			// if the message could not be deserialized, set $topic_instance_id to
			// false so that auto script creation does not get blocked
			$topic_instance_id = false;
		}
	}
	
	// if client id was provided, published the message, call the corresponding
	// topic script (if there is one)
	if($client_id !== false) {
		// Automatic topic script creation should happen only if the message is not
		// retained and there is no shadow variable / topic instance present!
		$allow_create = !$retain &&
			$topic_var_info['shadow_var_id'] === false &&
			$topic_instance_id === false;
		
		$topic_script_ids = get_topic_script_ids($topic_name, $allow_create);
		if($topic_script_ids !== false) {
			foreach($topic_script_ids as $topic_script_id) {
				$arr = array(
					'SENDER' => 'MQTT',
					'MESSAGE' => $message,
					'TOPIC' => $topic_name,
					'RETAIN' => $retain,
					'CLIENT' => $client_id
				);
				IPS_RunScriptEx($topic_script_id, $arr);
			}
		}
	}
		
	// messages without retain flag get published directly, while retained messages
	// will have to get published indirectly via the variable update events! This
	// ensures that updating variables in the topics category by other means than
	// mqtt clients will also get the changes published TO mqtt clients!
	if(!$retain) {
		send_publish_packages($topic_name, $message, $qos_override);
	}
} // publish

// populate a topic instance using the message, which is assumed to contain
// serialized data (json)
function populate_topic_instance($instance_id, $message) {
	$array = @json_decode($message, true);
	
	if(is_array($array)) {
		extract_array_to_ips($instance_id, $array);
		return true;
	} else {
		return false;
	}
} // populate_topic_instance

// Convert the contents of an array into a structure of ips categories and
// variables
function extract_array_to_ips($parent_id, $array) {
	foreach($array as $key => $value) {
		if(is_array($value)) {
			$subcat_id = @IPS_GetCategoryIDByName($key, $parent_id);
			if($subcat_id === false) {
				$subcat_id = IPS_CreateCategory();
				IPS_SetName($subcat_id, $key);
				IPS_SetParent($subcat_id, $parent_id);
			}
			extract_array_to_ips($subcat_id, $value);
		} else {
			$var_id = @IPS_GetVariableIDByName($key, $parent_id);
			if($var_id === false) {
				$ips_var_type = 3;
				if(is_bool($value)) {
					$ips_var_type = 0;
				} else if(is_int($value)) {
					$ips_var_type = 1;
				} else if(is_float($value)) {
					$ips_var_type = 2;
				}
				$var_id = IPS_CreateVariable($ips_var_type);
				IPS_SetName($var_id, $key);
				IPS_SetParent($var_id, $parent_id);
			}
			SetValue($var_id, $value);
		}
	}
} // extract_array_to_ips

// get or create a topic instance (dummy instance) to be populated with
// unserialized data from the topics' messages
function get_topic_instance_id($topic_name, $allow_create) {
	global $dummy_module_id;
	$cat_id = get_topic_cat_id($topic_name);
	$instance_id = @IPS_GetInstanceIDByName($topic_name, $cat_id);
	if($instance_id !== false) {
		$instance = IPS_GetInstance($instance_id);
		if($instance['ModuleInfo']['ModuleID'] != $dummy_module_id) {
			$instance_id = false;
		}
	}
	if($instance_id === false && $allow_create) {
		$instance_id = IPS_CreateInstance($dummy_module_id);
		IPS_SetName($instance_id, $topic_name);
		IPS_SetParent($instance_id, $cat_id);
	}
	return $instance_id;
} // get_topic_instance_id

// return array of script ids for
// - any wildcard scripts (named "#") along the way in the hierarchy
// - the script under the exact specified topic name, if it exists
// If $allow_create is false and no script (including wildcard scripts) exists, a
// new script is created with a default content, and only its id returned.
function get_topic_script_ids($topic_name, $allow_create) {
	// this variable will hold list of wildcard script ids after call to
	// get_topic_cat_id, or false
	$wildcard_script_ids = false;
	
	// will hold the end result returned by this function
	$result = false;
	
	// get the category id that the topic script might reside in. Also, populate
	// the wildcard script ids variable passed here
	$cat_id = get_topic_cat_id($topic_name, false, $wildcard_script_ids);
	
	// set result variable to contain any found wildcard script ids
	if($wildcard_script_ids !== false) {
		$result = $wildcard_script_ids;
	}
	
	// try to get the id of the topic script if it exists
	$script_id = false;
	if($cat_id !== false) {
		$script_id = @IPS_GetScriptIDByName($topic_name, $cat_id);
		
		// if no script was found, try if a link with the right name exists to a script
		if($script_id === false) {
			$link_id = @IPS_GetLinkIDByName($topic_name, $cat_id);
			if($link_id !== false) {
				$target_id = IPS_GetLink($link_id)['TargetID'];
				$obj = IPS_GetObject($target_id);
				if($obj['ObjectType'] == 3) { // it target is a script
					$script_id = $target_id;
				}
			}
		}
				
		// create topic script if none exists, but there is no need to create script
		// - for a retained message, as that will be preferrably handled by a variable
		// - if one or more wildcard scripts have been found on the category levels
		//   above
		if($script_id === false && $allow_create && $wildcard_script_ids === false) {
			$script_id = IPS_CreateScript(0);
			IPS_SetParent($script_id, $cat_id);
			IPS_SetName($script_id, $topic_name);
			// provide some dummy content for the script
			$content = '<?' . "\r" .
				'if($_IPS["SENDER"] == "MQTT") {' . "\r" .
				'  // $_IPS["MESSAGE"] contains the message' . "\r" .
				'  // $_IPS["TOPIC"] contains the topic' . "\r" .
				'  // $_IPS["CLIENT"] contains the client ID' . "\r" .
				'  // $_IPS["RETAIN"] is true or false' . "\r" .
				"\r  echo \"Unpopulated MQTT topic handler script called.\\r\";\r" .
				"  print_r(" . '$_IPS' . ");\r}\r" .
				"?>";
			IPS_SetScriptContent($script_id, $content);
		}
	}
	
	// if a topic script has been found, append it to the result array
	if($script_id !== false) {
		if($result === false) $result = array(); // create result array if necessary
		$result[] = $script_id;
	}
	
	return $result;
} // get_topic_script_ids

// Determines the variable profile for the given message. If the specified var_id
// is valid, the profile name of that variable is investigated and, if it fits the
// message, it is returned. Otherwise, a suitable variable profile is looked for
// and if one is found, its name is returned. If no profile is found, an empty
// string gets returned. If a variable profile name is returned, the $message
// also gets decoded into a raw value using that profile.
function get_var_profile(&$message, $var_id, &$guessed) {
	$ips_var_type = false;
	$prefix = '';
	$suffix = '';
	
	// first check whether the the message might be a date / time expression
	$time_value = strtotime($message);
	if($time_value !== false) {
		$ips_var_type = 1;
		$value = $time_value;
	} else {
		// try to split up the message in a prefix, a value and a suffix part
		$value = '';
		for($i = 0; $i < strlen($message); $i++) {
			$c = substr($message, $i, 1);
			if($c == '.' || $c == ',') {
				$value .= '.';
				$ips_var_type = 2; // value is float
			} else if(is_numeric($c)) {
				$value .= $c;
				if($ips_var_type != 2) $ips_var_type = 1;
			} else if($value == '') {
				$prefix .= $c;
			} else {
				$suffix .= $c;
			}
		}
	}
	
	$guessed = false;
	
	// if a valid variable id is specified, check out that variable's profile
	if($var_id !== false) {
		$var = @IPS_GetVariable($var_id);
		if($var !== false) {
			// string profiles are ignored as they offer nothing useful for us
			if($var['VariableType'] == 3) return '';
			
			// get the current profile of the variable
			$profile_name = $var['VariableCustomProfile'];
			if($profile_name == '') {
				$profile_name = $var['VariableProfile'];
			}
			
			if($profile_name != '') {
				$prof = @IPS_GetVariableProfile($profile_name);
				if($prof !== false) { // if variable has valid profile
					if(
						$profile_name == '~UnixTimestamp' ||
						$profile_name == '~UnixTimestampTime' ||
						$profile_name == '~UnixTimestampDate'
					) { // timestamp
						$message = sane_string_cast($value);
						return $profile_name;
					} else if(count($prof['Associations']) > 0) { // uses associations?
						foreach($prof['Associations'] as $assoc) {
							if($assoc['Name'] == $message) {
								$message = $assoc['Value'];
								return $profile_name;
							}
						}
					} else if($ips_var_type !== false) { // does not use associations
						if($prof['ProfileType'] == 1) $value = intval($value);
						if($prof['ProfileType'] == 2) $value = floatval($value);
						
						if(
							$value >= $prof['MinValue'] &&
							$value <= $prof['MaxValue'] &&
							$ips_var_type <= $prof['ProfileType']
						) {
							$message = sane_string_cast($value);
							return $profile_name;
						}
					}
				}
			}
		}
	} // if
	
	if($time_value !== false) { // timestamp
		$message = sane_string_cast($value);
		$guessed = true;
		return '~UnixTimestamp';
	} else if($prefix != '' || $suffix != '') {
		if($ips_var_type !== false) { // profile with prefix and/or suffix
			if($ips_var_type == 2) $value = floatval($value);
			if($ips_var_type == 1) $value = intval($value);
			for($vt = $ips_var_type; $vt <= 2; $vt++) {
				$profiles = IPS_GetVariableProfileListByType($vt);
				foreach($profiles as $profile_name) {
					$prof = IPS_GetVariableProfile($profile_name);
					if(
						(trim($prefix) == trim($prof['Prefix'])) &&
						(trim($suffix) == trim($prof['Suffix'])) &&
						(count($prof['Associations']) == 0) &&
						$value >= $prof['MinValue'] &&
						$value <= $prof['MaxValue']
					) {
						if($vt == 2) $value = floatval($value);
						$message = sane_string_cast($value);
						$guessed = true;
						return $profile_name;
					}
				}
			}
		} else { // profile with associations
			$assoc_name = $prefix . $suffix;
			
			$profiles = IPS_GetVariableProfileList();
			foreach($profiles as $profile_name) {
				$prof = IPS_GetVariableProfile($profile_name);
				foreach($prof['Associations'] as $assoc) {
					if($assoc['Name'] == $assoc_name) {
						$message = sane_string_cast($assoc['Value']);
						$guessed = true;
						return $profile_name;
					}
				}
			} // foreach
		} // else
	} // if
	
	return '';
} // get_var_profile

// returns the id of a link residing in the tree at the position specified by
// topic_name
function get_topic_link_id($topic_name) {
	$cat_id = get_topic_cat_id($topic_name);
	if($cat_id !== false) {
		return @IPS_GetLinkIDByName($topic_name, $cat_id);
	}
	return false;
} // get_topic_link_id

// returns an assoc array containing the variable ids for a topic's retained
// message
function get_topic_var_info($topic_name, $message, $create_now) {
	$flags = get_user_flags($topic_name);
	
	$cat_id = get_topic_cat_id($topic_name);
	
	$result = array(
		'default_var_id' => false,
		'default_var_is_linked' => false,
		'shadow_var_id' => false,
		'vestigial_var_ids' => array(
			'bool' => false,
			'int' => false,
			'float' => false
		)
	);
	
	$var_id = false;
	
	if($cat_id !== false && substr($topic_name, 0, 1) != '*') {
		$var_id = @IPS_GetVariableIDByName($topic_name, $cat_id);
		$result['shadow_var_id'] = @IPS_GetVariableIDByName('*' . $topic_name, $cat_id);
		
		if($flags & USE_VAR_PROFILE) {
			$profile_name = get_var_profile($message, $var_id, $guessed);
			debug_log(
				'Profile' . ($guessed ? ' (guessed)' : '') . ': "' . $profile_name . '"' .
				', decoded raw value of message: ' . $message);
		}
		
		// if no variable was found, try if a link with the right name exists to a variable
		if($var_id === false) {
			$link_id = @IPS_GetLinkIDByName($topic_name, $cat_id);
			if($link_id !== false) {
				$target_id = IPS_GetLink($link_id)['TargetID'];
				$obj = IPS_GetObject($target_id);
				if($obj['ObjectType'] == 2) { // if target is a variable
					$result['default_var_is_linked'] = true;
					$var_id = $target_id;
				}
			}
		}
		
		// we try to guess the variable type. If we're guessing wrong, it's preferrable
		// to manually create a variable in the right place with the right name of
		// the desired data type...
		$ips_var_type = false;
		if(is_numeric($message)) {
			if(strpos($message, '.') !== false) {
				$ips_var_type = 2;
			} else if(intval($message) != 0 || $message == '0') {
				$ips_var_type = 1;
			} else {
				debug_log("Numeric value not basic: \"" .  $message . '"' .
					", intval: \"" . sane_string_cast(intval($message)) . "\", floatval: \"" .
					sane_string_cast(floatval($message)) . "\"");
				$ips_var_type = 3;
			}
		} else if($message == 'false' || $message == 'true') {
			$ips_var_type = 0;
		} else if($message != '') {
			$ips_var_type = 3;
		}
		//IPS_LogMessage('MQTT', $message . ": " . $ips_var_type);
		
		// Type escalation:
		// Check if the type of an existing variable would be unable to hold the
		// new message. If not, create a new variable of a suitable type end rename
		// the old one (can be disabled in user config)
		// Will not be performed for linked variables.
		if(($flags & NO_TYPE_ESCALATION) == 0 && !$result['default_var_is_linked']) {
			if($var_id !== false && $ips_var_type !== false) {
				$var = IPS_GetVariable($var_id);
				if($var['VariableType'] < $ips_var_type) {
					if($var['VariableType'] == 0) $vt_suffix = 'bool';
					if($var['VariableType'] == 1) $vt_suffix = 'int';
					if($var['VariableType'] == 2) $vt_suffix = 'float';
					IPS_SetName(
						$var_id,
						$topic_name . '@' . $vt_suffix
					);
					ensure_var_event($var_id);
					$var_id = false;
				}
			}
			
			// gather the ids of potential "vestigial" variables left over from type
			// escalation
			$result['vestigial_var_ids'] = array(
				'bool' => @IPS_GetVariableIDByName($topic_name . '@bool', $cat_id),
				'int' => @IPS_GetVariableIDByName($topic_name . '@int', $cat_id),
				'float' => @IPS_GetVariableIDByName($topic_name . '@float', $cat_id)
			);
		}
		
		if($create_now) {
			if($var_id === false) {
				if($ips_var_type !== false) {
					$var_id = IPS_CreateVariable($ips_var_type);
					IPS_SetParent($var_id, $cat_id);
					IPS_SetName($var_id, $topic_name);
				}
			}
		}
		ensure_var_event($var_id);
	}
	
	$result['default_var_id'] = $var_id;
	
	debug_log('Topic variable info: ' . print_r($result, true));
	
	return $result;
} // get_topic_var_info

// ensure that the variable with the provided id has an "on update" event which
// calls this script. Returns the event's id.
function ensure_var_event($var_id) {
	global $event_ids, $events;
	
	// if not cached already, get list of event ids for this script
	if($event_ids === false) $event_ids = IPS_GetChildrenIDs($_IPS['SELF']);
	
	// get the event or retrieve it from cache
	foreach($event_ids as $event_id) {
		if(array_key_exists($event_id, $events)) {
			$event = $events[$event_id];
		} else {
			$event = IPS_GetEvent($event_id);
			$events[$event_id] = $event;
		}
		
		// if event exists for the supplied var id, return it
		if($event['TriggerVariableID'] == $var_id) {
			return $event_id;
		}
	}
	
	$event_id = IPS_CreateEvent(0);
	IPS_SetEventTrigger($event_id, 0, $var_id);
	IPS_SetParent($event_id, $_IPS['SELF']);
	// get the topic name - will only work for variables that reside within the
	// topics structure (not for linked variables that reside elsewhere!)
	$topic_name = get_topic_from_id($var_id);
	if($topic_name != '') { // only set the event name to the topic name if we know it
		IPS_SetName($event_id, $topic_name);
	}
	IPS_SetEventActive($event_id, true);
	$event_ids[] = $event_id;
	return $event_id;
} // ensure_var_event

// returns the id of the immediate category in which the topic variable/script
// for the specified topic name resides. The specified $topic_name gets changed
// to only contain the last part of the topic, i.e. the name of the variable or
// script. The returned category will likely be part of a nested structure of
// categories, which lies below the $topics_cat_id (topic root)
function get_topic_cat_id(&$topic_name, $parent_cat_id = false, &$wildcard_script_ids = false) {
	global $topics_cat_id;
	if($topics_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Topics Root category not configured - run setup!');
	}
	
	if($parent_cat_id === false) $parent_cat_id = $topics_cat_id;
	
	// check whether a wildcard script exists at this topic level, and if
	// so, enter it into the wildcard script ids list
	$wildcard_script_id = @IPS_GetScriptIDByName('#', $parent_cat_id);
	
	// if no script of the wildcard name exists, try if perhaps a link of the
	// name exists to a script
	if($wildcard_script_id === false) {
		$link_id = @IPS_GetLinkIDByName('#', $parent_cat_id);
		if($link_id !== false) {
			$target_id = IPS_GetLink($link_id)['TargetID'];
			$obj = @IPS_GetObject($target_id);
			if($obj !== false) {
				if($obj['ObjectType'] == 3) {
					IPS_SetPosition($link_id, -1);
					$wildcard_script_id = $target_id;
				}
			}
		}
	} else {
		IPS_SetPosition($wildcard_script_id, -1);
	}
	
	// if wildcard script found, add it to the wildcard script ids list
	if($wildcard_script_id !== false) {
		if($wildcard_script_ids === false) $wildcard_script_ids = array();
		$wildcard_script_ids[] = $wildcard_script_id;
	}
	
	$slash_pos = strpos($topic_name, '/');
	
	// if there is a slash in the remaining string, there is a sub topic, so we
	// must still be in a "category" part of the structure
	if($slash_pos !== false) {
		// category levels without a name will be substituted by a single slash!
		if($slash_pos > 0) { // if there is a name supplied for the current topic level
			$cat_name = substr($topic_name, 0, $slash_pos); // get the topic level's name
		} else { // if the level is unnamed (e.g. initial slash used)
			$cat_name = '/'; // substitute the empty name for a single slash
		}
		$topic_name = substr($topic_name, $slash_pos + 1); // set to remainder of name
				
		if($topic_name == "") $topic_name = '/'; // if remainder is empty, set to single slash instead
		
		// get the category id and if the category does not exist, create it
		$cat_id = @IPS_GetCategoryIDByName($cat_name, $parent_cat_id);
		if($cat_id === false) {
			$cat_id = IPS_CreateCategory();
			IPS_SetParent($cat_id, $parent_cat_id);
			IPS_SetName($cat_id, $cat_name);
		}
		
		if(strpos($topic_name, '/') === false || $topic_name == '/') { // if there is no sub category
			return $cat_id;
		} else { // if there is a sub category, call this function recursively
			return get_topic_cat_id($topic_name, $cat_id, $wildcard_script_ids);
		}
	}
} // get_topic_cat_id

// converts a connection id to a string without special characters, which can
// be used as a variable ident and/or client id.
// 127.0.0.1:1000 => 127x0x0x1p1000
function conn_id_to_ident($conn_id) {
	return str_replace(array('.', ':'), array('x', 'p'), $conn_id);
}
function ident_to_conn_id($ident) { // other way round
	return str_replace(array('x', 'p'), array('.', ':'), $ident);
}

// return most significant byte of value, pretending it's a two byte integer
function msb($value) {
	return floor($value / 256);
}

// return least significant byte of value, pretending it's a two byte integer
function lsb($value) {
	return $value % 256;
}

// return assoc array of a client session specified by either the $conn_id or the
// $client_id. ($conn_id represents a tcp connection to the server socket,
// $client_id represents the client id of an mqtt session)
function get_session($conn_id, $client_id = false, $create_if_necessary = false, $clean = false) {
	global $sessions_cat_id;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	// Each session is stored in a string variable. The session variables all lie
	// under $sessions_cat_id. Below each session variable, one or more subscription
	// variables may lie.
	$session_var_id = @IPS_GetObjectIDByIdent(conn_id_to_ident($conn_id), $sessions_cat_id);
	
	if($session_var_id !== false) { // if session found through the connection id
		debug_log('Session variable for connection "' . $conn_id . '" retrieved. ID: ' . $session_var_id);
		$session_found_through_client_id = false;
		$session_found_through_connection_id = true;
	} else if($client_id !== false) { // otherwise, if client id was supplied
		$session_found_through_connection_id = false;
		$session_var_id = @IPS_GetVariableIDByName($client_id, $sessions_cat_id);
		if($session_var_id !== false) {
			debug_log('Session not found through connection data, variable retrieved ' .
				'via client ID "' . $client_id . '". Variable ID: ' . $session_var_id);
			$session_found_through_client_id = true;
		}
	}
	
	if($session_var_id !== false && $clean) {
		debug_log('Clean session flag set - need to delete session variable.');
		delete_session_variable($session_var_id);
		$session_var_id = false;
	}
	
	$session_assoc = false;
	
	if($session_var_id === false) { // if there was no session variable present, create one
		if($create_if_necessary) {
			debug_log('No session variable - need to create one.');
			$session_var_id = create_session_variable($client_id, $conn_id);
		} else {
			debug_log('No session variable!');
			return false; // no session and not allowed to create one
		}
	} else { // if there is already a session variable, read and parse it
		$session_str = GetValue($session_var_id);
		$session_assoc = @json_decode($session_str, true);
		
		if( // check for integrity of session object
			!is_array($session_assoc) ||
			!array_key_exists('last_update', $session_assoc) ||
			!array_key_exists('client_id', $session_assoc) ||
			!array_key_exists('variable_id', $session_assoc) ||
			!array_key_exists('clean', $session_assoc) ||
			!array_key_exists('new', $session_assoc) ||
			!array_key_exists('package_ids', $session_assoc) ||
			!is_array($session_assoc['package_ids']) ||
			!array_key_exists('outgoing_next', $session_assoc['package_ids']) ||
			!array_key_exists('incoming_received_qos_2', $session_assoc['package_ids']) ||
			!array_key_exists('outgoing_sent_qos_1', $session_assoc['package_ids']) ||
			!array_key_exists('outgoing_sent_qos_2', $session_assoc['package_ids']) ||
			!array_key_exists('outgoing_received_qos_2', $session_assoc['package_ids'])
		) {
			debug_log('Session corrupted:\r' . $session_str);
			delete_session_variable($session_var_id);
			$session_assoc = false; // discard session object if it seems corrupted
		}
		
		$kal = $session_assoc['kal'];
		$last_update = $session_assoc['last_update'];
		
		if($kal > 0 && time() > $last_update + $kal * 1.5 && $clean) { // if old "clean" session has timed out, discard it
			debug_log('Discarding session data retrieved from variable, because it ' .
				'holds an expired "clean" session.\rOld session data: ' .
				print_r($session_assoc, true));
			perform_last_will($session_assoc);
			$session_assoc = false;
		} else if($session_found_through_client_id && $session_assoc['connection_id'] !=
			$conn_id && $clean) { // if "clean" session started under a different tcp connection, discard old session
			debug_log('Discarding session data retrieved from variable, because it ' .
				'holds a "clean" session started over a different TCP connection.\r' .
				'Old session data: ' . print_r($session_assoc, true));
			$session_assoc = false;
		} else if($client_id && $session_found_through_connection_id &&
			$session_assoc['client_id'] != $client_id) { // if client ids inconsistent, create new session
			debug_log('Discarding session data retrieved from variable and starting' .
				' a new session with a new variable, because the cliend IDs did not ' .
				'match. (Current: "' . $client_id . '", from variable: "' .
				$session_assoc['client_id'] . '"');
			$session_var_id = create_session_variable($client_id, $conn_id);
			$session_assoc = false;
		}
	}
	
	// if there is no session data (new session, or old session discarded), create
	// a new session dataset
	if($session_assoc === false) {
		if($create_if_necessary) {
			$session_assoc = array(
				"established" => time(),
				"last_update" => time(),
				"client_id" => $client_id,
				"variable_id" => $session_var_id, // needed to update or end the session
				"clean" => $clean,
				"new" => true, // flag to indicate that the session was created just now
				"package_ids" => array(
					"outgoing_next" => 1, // next package id to be used for outgoing package with qos > 0
					"incoming_received_qos_2" => array(), // PT_PUBREC sent, awaiting PT_PUBREL
					"outgoing_sent_qos_1" => array(), // PT_PUBLISH sent, awaiting PT_PUBACK
					"outgoing_sent_qos_2" => array(), // PT_PUBLISH sent, awaiting PT_PUBREC
					"outgoing_received_qos_2" => array() // PT_PUBREL sent, awaiting PT_PUBCOMP
				)
			);
			debug_log('Session created!');
		} else {
			debug_log('No session data could be found.');
			return false;
		}
	} else {
		debug_log('Session data was found and will be reused.');
		$session_assoc["new"] = false;
	}
	
	// always update the connection id with that of the current connection before
	// the session gets used by any other function
	$session_assoc['connection_id'] = $conn_id;
	
	// update the ident as well so that the session can be found by looking for
	// the conn_id
	$ident = conn_id_to_ident($conn_id);
	if(!@IPS_SetIdent($session_var_id, $ident)) {
		$old_ident_var_id = IPS_GetObjectIDByIdent($ident, $sessions_cat_id);
		IPS_SetIdent($old_ident_var_id, '');
		IPS_SetHidden($old_ident_var_id, true);
		IPS_SetIdent($session_var_id, $ident);
	}
	IPS_SetHidden($session_var_id, false);
	
	debug_log('Session data: ' . print_r($session_assoc, true));
	return $session_assoc;
} // get_session

// creates the session variable with the specified parameters. Used internally
// by get_session.
function create_session_variable($client_id, $conn_id) {
	global $sessions_cat_id;
	if($sessions_cat_id === false) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		throw new Exception('Sessions category not configured - run setup!');
	}
	
	debug_log('Creating a session variable for connection "' . $conn_id .
		'", client ID "' . $client_id . '")');
	$session_var_id = IPS_CreateVariable(3);
	if(!@IPS_SetParent($session_var_id, $sessions_cat_id)) {
		IPS_SetScriptTimer($_IPS['SELF'], 0);
		IPS_DeleteVariable($session_var_id);
		throw new Exception('Sessions category has been deleted - run setup!');
	}
	
	// Temporary cleanup routine to remove unnecessary duplicate session variables
	// that have been created by previous, buggy versions of this script
	$count = 0;
	do {
		$count++;
		$existing_session_var_id = @IPS_GetVariableIDByName($client_id, $sessions_cat_id);
		if($existing_session_var_id !== false) {
			debug_log('A session variable for the same client ID already existed. ' .
				'This should not happen! Deleting existing variable to avoid pile-up.');
			IPS_DeleteVariable($existing_session_var_id);
		} else break;
	} while($count < 100); // only remove max. 100 at a time
	
	IPS_SetName($session_var_id, $client_id);
	$ident = conn_id_to_ident($conn_id);
	if(!@IPS_SetIdent($session_var_id, $ident)) {
		$old_ident_var_id = IPS_GetObjectIDByIdent($ident, $sessions_cat_id);
		debug_log('A session variable for the same connection already existed and' .
			' will be detached from that connection (ID: ' . $old_ident_var_id . ')');
		IPS_SetIdent($old_ident_var_id, '');
		IPS_SetHidden($old_ident_var_id, true);
		IPS_SetIdent($session_var_id, $ident);
	}
	IPS_SetHidden($session_var_id, false);
	
	debug_log('Session variable created. ID: ' . $session_var_id . ')');
	return $session_var_id;
} // create_session_variable

// Delete a session variable, including potential subscription variables (which
// would be children of the session variable)
function delete_session_variable($var_id) {
	debug_log('Deleting session variable (ID: ' . $var_id . ')');
	$sub_var_ids = IPS_GetChildrenIDs($var_id);
	foreach($sub_var_ids as $sub_var_id) {
		if(!@IPS_DeleteVariable($sub_var_id)) return false;
	}
	return(@IPS_DeleteVariable($var_id));
} // delete_session_variable

// updates a session's 'last_update' time stamp (used for the keep alive timeout)
// and then stores the session in its session variable
function update_session($session_assoc) {
	if($session_assoc !== false) {
		if(array_key_exists('variable_id', $session_assoc)) {
			$session_var_id = $session_assoc['variable_id'];
			if(IPS_ObjectExists($session_var_id)) {
				debug_log('Updating session (Variable ID: ' . $session_var_id . ').');
				$session_assoc['last_update'] = time();
				$session_assoc["new"] = false;
				$session_str = json_encode($session_assoc);
				debug_log('Session data: ' . print_r($session_assoc, true));
				SetValue($session_var_id, $session_str);
			} else {
				throw new Exception("Session variable missing unexpectedly: ID " . $session_var_id);
			}
		} else {
			debug_log('Unable to update session, because session data was empty.');
		}
	} else {
		debug_log('Unable to update session, because session data was empty.');
	}
} // update_session

// Ends the TCP session. For a persistent session that means only to discard the
// conn_id. For "clean" sessions, the session variable gets deleted entirely as
// there is no way to resume it.
function end_session($session_assoc) {
	if($session_assoc !== false) {
		$session_var_id = $session_assoc['variable_id'];
		if(IPS_ObjectExists($session_var_id)) {
			if($session_assoc['clean']) { // if non persistent session...
				// ...delete session entirely
				if(!delete_session_variable($session_var_id)) {
					throw new Exception('Unable to delete session variable!');
				}
				debug_log('Session variable (ID: ' . $session_var_id . ') deleted.');
			} else { // if persistent session...
				IPS_SetHidden($session_var_id, true);
				IPS_SetIdent($session_var_id, ''); // ...clear the ident (i.e. conn_id)
				debug_log('Session variable (ID: ' . $session_var_id . ') detached from connection.');
			}
		}
	}
} // end_session

// send an mqtt package over the specified tcp connection
function send_package($conn_id, $type, $flags, $bytes = false) {
	global $regvar_id, $pt_names, $debug;
	
	debug_log("Sending over " . $conn_id . ": " . $pt_names[$type] . "...");
	
	// get ip and port from conn_id
	$colon_pos = strpos($conn_id, ':');
	$ip = substr($conn_id, 0, $colon_pos);
	$port = substr($conn_id, $colon_pos + 1);
	
	// if no bytes supplied, create empty bytes array
	if($bytes === false) $bytes = array();
	
	// add fixed header at beginning of bytes array
	array_unshift($bytes, count($bytes));
	array_unshift($bytes, ($type << 4) | $flags);
	
	// create bytes string from bytes array
	$data = "";
	for($i = 0; $i < count($bytes); $i++) {
		if($debug && !is_numeric($bytes[$i])) {
			debug_log("Encoding warning: Value for byte " . $i . " does not look well formed! \"" . $bytes[$i] . "\"!");
		}
		$data .= chr($bytes[$i]);
	}
	
	// send and return true if sending succeeded
	return @RegVar_SendPacket($regvar_id, $data, $ip, intval($port));
}

// extract and return the (package type specific) payload, or false if more data
// is expected
function extract_payload(&$header, &$data) {
	global $pt_names;
	$payload = array();
	
	$remaining_len = $header['remaining_len'];
	
	switch($header['type']) {
		case PT_CONNECT:
			$str = extract_string($data, $remaining_len);
			if($str === false) return false;
			$payload['client_id'] = $str;
			
			if($header['connect_flags']['will_flag']) {
				$str = extract_string($data, $remaining_len);
				if($str === false) return false;
				$payload['will_topic'] = $str;
				
				$str = extract_binary($data, $remaining_len);
				if($str === false) return false;
				$payload['will_message'] = $str; // TODO: check why this would not work with mqtt-spy
			}
			
			if($header['connect_flags']['user_name']) {
				$str = extract_string($data, $remaining_len);
				if($str === false) return false;
				$payload['user_name'] = $str;
			}
			
			if($header['connect_flags']['password']) {
				$str = extract_binary($data, $remaining_len);
				if($str === false) return false;
				$payload['password'] = $str;
			}
			break;
		case PT_PUBLISH:
			if(strlen($data) < $remaining_len) return false;
			$payload = substr($data, 0, $remaining_len);
			$data = substr($data, $remaining_len);
			$remaining_len = 0;
			break;
		case PT_SUBSCRIBE:
			while($remaining_len > 0) {
				$topic = extract_string($data, $remaining_len);
				if($topic === false) return false;
				
				if(strlen($data) < 1) return false;
				
				$qos = ord($data[0]);
				if($qos > 2) throw new Exceptions("Unsupported qos level for subscription. Must be 0..2");
				$data = substr($data, 1);
				$remaining_len--;
				
				$payload[] = array('topic_filter' => $topic, 'qos' => $qos);
			}
			break;
		case PT_UNSUBSCRIBE:
			while($remaining_len > 0) {
				$topic = extract_string($data, $remaining_len);
				if($topic === false) return false;
				
				$payload[] = $topic;
			}
			break;
		default:
			throw new Exception('Unexpected payload received for package type ' . $pt_names[$header['type']]);
	}
	
	$header['remaining_len'] = $remaining_len;
	return $payload;
} // extract_payload

// extract header from mqtt package
function extract_header(&$data) {
	global $pt_names;
	
	if(strlen($data) < 2) return false;
	$byte = ord($data[0]);
	
	$type = ($byte & 0b11110000) >> 4;
	$flags = $byte & 0b00001111;
	
	$expect_topic_name = false;
	$expect_package_id = false;
	
	// check for validity of package type and flags
	switch($type) {
		case PT_RESERVED_0:
		case PT_RESERVED_15:
			throw new Exception('Package type not allowed: ' . $pt_names[$type]);
		case PT_CONNACK:
		case PT_SUBACK:
		case PT_UNSUBACK:
		case PT_PINGRESP:
			throw new Exception('Package type not allowed for c2s communication: ' . $pt_names[$type]);
		case PT_PUBACK:
		case PT_PUBREC:
		case PT_PUBCOMP:
			$expect_package_id = true;
		case PT_DISCONNECT:
		case PT_PINGREQ:
		case PT_CONNECT:
			if($flags == 0b00000000) break;
			throw new Exception('Invalid package flags for ' . $pt_names[$type] . ' package.');
		case PT_PUBLISH:
			$flags = array( // reformat flags into a neat assoc array
				"duplicate" => boolval($flags & 0b00001000),
				"qos" => ($flags & 0b00000110) >> 1,
				"retain" => boolval($flags & 0b00000001)
			);
			$expect_topic_name = true;
			$expect_package_id = $flags['qos'] > 0;
			break;
		case PT_PUBREL:
		case PT_SUBSCRIBE:
		case PT_UNSUBSCRIBE:
			$expect_package_id = true;
			if($flags == 0b00000010) break;
			throw new Exception('Invalid package flags for ' . $pt_names[$type] . ' package.');
	}
	
	// shift data buffer so that the next field can be extracted
	$data = substr($data, 1);
	
	// fixed header fields are always present
	$header = array(
		"type" => $type,
		"flags" => $flags,
	);
	
	// extract remaining length field
	$remaining_len = extract_remaining_length($data);
	if($remaining_len === false) return false; // data ends prematurely
	
	if($expect_topic_name) {
		$topic_name = extract_string($data, $remaining_len);
		if($topic_name === false) return false; // data ends prematurely
		$header['topic_name'] = $topic_name;
	}
	
	// package id is part of variable header and only expected for certain package types
	if($expect_package_id) {
		if(strlen($data) < 2) return false;
		
		if($remaining_len < 2) {
			throw new Exception('Remaining package length unexpectedly short (Package id of 2 bytes expected)');
		}
		$msb = ord($data[0]);
		$lsb = ord($data[1]);
		$package_id = $msb * 256 + $lsb;
		
		$data = substr($data, 2);
		$remaining_len -= 2;
		if($package_id == 0) throw new Exception('Invalid package id: Must be nonzero.');
		$header["package_id"] = $package_id;
	}
		
	// additional variable header fields depending on header type
	switch($type) {
		case PT_CONNECT:
			$protocol_name = extract_string($data, $remaining_len);
			if($protocol_name === false) return false;
			if(strtoupper($protocol_name) != "MQTT" && strtoupper($protocol_name) != "MQISDP") {
				throw new Exception('Unrecognized protocol name: ' . $protocol_name);
			}
			
			if(strlen($data) < 4) return false;
			if($remaining_len < 4) {
				throw new Exception('Remaining length unexpectedly short (Variable header of at least 4 more bytes expected)');
			}
			$header['protocol_name'] = $protocol_name;
			$protocol_level = ord($data[0]);
			//if($protocol_level != 4) throw new Exception('Unsupported protocol level (' . $protocol_level . ', should be 4)');
			$header['protocol_level'] = $protocol_level;
			$connect_flags = ord($data[1]);
			if($connect_flags & 0b00000001) throw new Exception('Reserved flag byte 0 of connect flags must be zero.');
			$connect_flags = array(
				"user_name" => boolval($connect_flags & 0b10000000),
				"password" => boolval($connect_flags & 0b01000000),
				"will_retain" => boolval($connect_flags & 0b00100000),
				"will_qos" => ($connect_flags & 0b00011000) >> 3,
				"will_flag" => boolval($connect_flags & 0b0000100),
				"clean_session" => boolval($connect_flags & 0b0000010),
			);
			$header['connect_flags'] = $connect_flags;
			$kal_msb = ord($data[2]);
			$kal_lsb = ord($data[3]);
			$kal = $kal_msb * 256 + $kal_lsb;
			$header['kal'] = $kal;
			
			$data = substr($data, 4);
			$remaining_len -= 4;
			break;
	}
	
	$header["remaining_len"] = $remaining_len;
	
	return $header;
} // extract_header

// extract the "remaining length" field and return its value as an integer
function extract_remaining_length(&$data) {
	$result = 0;
	$multi = 1;
	do {
		if(strlen($data) < 1) return false;
		
		$byte = ord($data[0]);
		$data = substr($data, 1); // shift data buffer
		
		$result += ($byte & 0b01111111) * $multi;
		
		$multi *= 128;
		
		if($multi > 128 * 128 * 128) throw new Exception('Malformed remaining length (MSBit of fourth byte must be zero)');
	} while($byte & 0b10000000);
	
	return $result;
} // extract_remaining_length

// extract UTF-8 string data from mqtt packet and return it
function extract_string(&$data, &$remaining_len) {
	if(strlen($data) < 2) return false;
	if($remaining_len < 2) throw new Exception('Remaining length unexpectedly short (UTF-8 string length field of 2 bytes expected)');
	$len_msb = ord($data[0]);
	$len_lsb = ord($data[1]);
	$len = $len_msb * 256 + $len_lsb;
	if(strlen($data) < $len + 2) return false;
	$result = substr($data, 2, $len);
	$data = substr($data, 2 + $len);
	$remaining_len -= 2 + $len;
	if(strpos($result, chr(0)) !== false) throw new Exception('Null character not allowed in UTF-8 string!');
	if(!mb_check_encoding($result, 'UTF-8')) throw new Exception('Extracted string encoding is not valid UTF-8!');
	return $result;
}

function extract_binary(&$data, &$remaining_len) {
	if(strlen($data) < 2) return false;
	if($remaining_len < 2) throw new Exception('Remaining length unexpectedly short (length field of 2 bytes expected)');
	$len_msb = ord($data[0]);
	$len_lsb = ord($data[1]);
	$len = $len_msb * 256 + $len_lsb;
	if(strlen($data) < $len + 2) return false;
	$result = substr($data, 2, $len);
	$data = substr($data, 2 + $len);
	$remaining_len -= 2 + $len;
	return preg_replace('/[[:^print:]]/', "", $result); // remove nonprintable characters in order to not break IPS database
}

function debug_log($text, $alert = false) {
	global $debug;
	if($debug || $alert) {
		if($_IPS['SENDER'] == 'Execute') {
			echo $text . "\r";
		} else {
			IPS_LogMessage('MQTT Server', $text);
		}
	}
}

// just a plain old "to string" cast, like it exists in any other language,
// without such php-specific nonsense as unwanted localized of float numbers
function sane_string_cast($value) {
	if(is_float($value)) { // jump through hoops to get a number in 0.005 format
		$current = $value - floor($value);
		for($decimals = 0; ceil($current); $decimals++) {
		    $current = ($value * pow(10, $decimals + 1)) - floor($value * pow(10, $decimals + 1));
		}
		return number_format($value, $decimals, '.', '');
	} else {
		return (string)$value;
	}
}
?>